/* arg.cc
 *
 * Copyright 2004,2005 Chun-Chung Chen <cjj@u.washington.edu>
 *
 * This program is free software; you can redistribute it and'or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Revision: 2005-03-18
 *  - add SetValue
 */

#include <iostream>
#include <arg.hh>
#include <cctype>

using namespace arg;
using namespace std;

Value::~Value()
{
}

Option::Option(int k, string n) :
	store_ptr(0),
	store_optional(false),
	set_var(0),
	set_once(false),
	call_func(0)
{
	key = k;
	name = n;
}

Option::~Option()
{
	if (store_ptr) {
		delete store_ptr;
	}
}

Option & Option::store(Value * ptr)
{
	store_ptr = ptr;
	return * this;
}

Option & Option::optional(string str)
{
	store_optional = true;
	store_str = str;
	return * this;
}

Option & Option::set(int * var, int value)
{
	set_var = var;
	set_value = value;
	return * this;
}

Option & Option::once(int init)
{
	set_once = true;
	set_init = init;
	return * this;
}

Option & Option::call(CallBack * func, void * data)
{
	call_func = func;
	call_data = data;
	return * this;
}

Option & Option::help(string text, string var)
{
	help_text = text;
	help_var = var;
	return * this;
}

bool Option::take_value()
{
	return store_ptr;
}

bool Option::need_value()
{
	return store_ptr && ! store_optional;
}

int Option::get_key()
{
	return key;
}

string Option::get_name()
{
	return name;
}

string Option::get_help(HelpFormat format)
{
	string h;
	bool s = isprint(key) && ! isspace(key);

	switch (format) {
	case HF_REGULAR:
		h = s ? (string("  -") + char(key)) : "    ";
		if (name != "") h += (s ? ", --" : "  --") + name;
		if (store_ptr) {
			if (store_optional) h += "[=" + help_var + "]";
			else h += "=" + help_var;
		}
		if (h.size() < 26) h.resize(26, ' ');
		h += "   ";
		h += help_text;
		break;
	case HF_NODASH:
		h = "    ";
		if (s) h += char(key);
		if (name != "") h += (s ? ", " : "") + name;
		if (store_ptr) {
			if (store_optional) h += "[=" + help_var + "]";
			else h += "=" + help_var;
		}
		if (h.size() < 26) h.resize(26, ' ');
		h += "   ";
		h += help_text;
		break;
	}
	return h;
}

void Option::process()
{
	if (store_ptr) {
		if (! store_optional) throw OptError(name, "missing value");
		store_ptr->set(store_str);
	}
	if (set_var) {
		if (set_once && set_init != * set_var) throw OptError(name, "can not re-set");
		* set_var = set_value;
	}
	if (call_func) {
		if (! (* call_func)(key, "", call_data)) throw OptError(name, "callback error");
	}
}

void Option::process(string str)
{
	if (! store_ptr) throw OptError(name, "unwanted value");
	store_ptr->set(str);
	if (set_var) {
		if (set_once && set_init != * set_var) throw OptError(name, "can not re-set");
		* set_var = set_value;
	}
	if (call_func) {
		if (! (* call_func)(key, str, call_data)) throw OptError(name, "callback error");
	}
}

Parser::~Parser()
{
	while (opt_list.size())	{
		delete opt_list.back();
		opt_list.pop_back();
	}
}

void Parser::add_help(string msg)
{
	HelpLine hl;
	hl.msg = msg;
	hl.opt = 0;
	help_list.push_back(hl);
}

Option & Parser::add_opt(int key, string name, bool hide)
{
	Option * o = new Option(key, name);
	opt_list.push_back(o);
	if (! hide) {
		HelpLine hl;
		hl.msg = "";
		hl.opt = o;
		help_list.push_back(hl);
	}
	return * o;
}

vector<string> & Parser::args()
{
	return arg_list;
}

bool Parser::parse(int argc, char * argv[])
{
	arg_list.clear();
	for (int i = 0; i < argc; i ++) {
		string s = argv[i];
		if (s[0] != '-') { // non-option => argument
			arg_list.push_back(s);
			continue;
		}
		if (s[1] == '-') { // long options
			string n; // name
			string v; // value
			bool vp = false; // value exists?
			unsigned k = s.find('=');
			if (k != string::npos) { // value present
				n = s.substr(2, k - 2);
				v = s.substr(k + 1);
				vp = true;
			} else {
				n = s.substr(2);
			}
			vector<Option *>::iterator j = opt_list.begin();
			while (true) {
				if (j == opt_list.end()) throw UnknError(n);
				if ((* j)->get_name() == n) break;
				j ++;
			}
			if (vp) (* j)->process(v);
			else (* j)->process();
			continue;
		}
		// short options
		for (unsigned k = 1; k < s.length(); k ++) {
			vector<Option *>::iterator j = opt_list.begin();
			while (true) {
				if (j == opt_list.end()) throw UnknError(string("-") + s[k]);
				if ((* j)->get_key() == s[k]) break;
				j ++;
			}
			if (! (* j)->take_value()) { // no value allowed
				(* j)->process();
				continue;
			}
			// value allowed, it could follow
			string v = s.substr(k + 1);
			if (v != "") {
				(* j)->process(v);
				break;
			}
			if (! (* j)->need_value() || ++ i >= argc) {
				(* j)->process();
				break;
			} else (* j)->process(argv[i]);
			break;
		}
	}
	return true;
}

string Parser::get_help()
{
	string h;
	for (vector<HelpLine>::iterator i = help_list.begin(); i != help_list.end(); i ++) {
		h += i->msg;
		if (i->opt) {
			h += i->opt->get_help();
		}
		h += '\n';
	}
	return h;
}



namespace { // local callback functions
	bool help_callback(int key, string str, void * data)
	{
		Parser * p = static_cast<Parser *>(data);
		cout << p->get_help();
		exit(0);
	}

	bool version_callback(int key, string str, void * data)
	{
		string * s = static_cast<string *>(data);
		cout << * s << '\n';
		exit(0);
	}
}

Option & Parser::add_opt_help()
{
	return add_opt('?', "help")
		.call(& help_callback, this)
		.help("display this help list and exit");
}

Option & Parser::add_opt_version(string ver)
{
	version_info = ver;
	return add_opt('V', "version")
		.call(& version_callback, & version_info)
		.help("print program version and exit");
}

Option & SubParser::add_opt(std::string name)
{
	return Parser::add_opt(0, name);
}

void SubParser::set(string str)
{
	string name;
	string value;
	int s = 0;
	unsigned n = string::npos;
	for (unsigned k = 0; k < str.length(); k ++) {
		if (str[k] == '=') n = k;
		if (k + 1 == str.length() || str[k + 1] == ',') {
			if (n == string::npos) n = k + 1;
			name = str.substr(s, n - s);
			vector<Option *>::iterator j = opt_list.begin();
			while (true) {
				if (j == opt_list.end()) throw UnknError(name);
				if ((* j)->get_name() == name) break;
				j ++;
			}
			if (n < k + 1) {
				value = str.substr(n + 1, k - n);
				(* j)->process(value);
			} else (* j)->process();
			k ++;
			s = k + 1;
		}
	}
}

string SubParser::get_help()
{
	string h;
	for (vector<HelpLine>::iterator i = help_list.begin(); i != help_list.end(); i ++) {
		h += i->msg;
		if (i->opt) {
			h += i->opt->get_help(Option::HF_NODASH);
		}
		h += '\n';
	}
	return h;
}

namespace { // local callback functions
	bool sub_help_callback(int key, string str, void * data)
	{
		SubParser * p = static_cast<SubParser *>(data);
		cout << p->get_help();
		exit(0);
	}
}

Option & SubParser::add_opt_help()
{
	return add_opt("help")
		.call(& sub_help_callback, this)
		.help("display this help list and exit");
}

Error::Error()
{
}

Error::Error(string m)
{
	msg = m;
}

string Error::get_msg()
{
	return msg;
}

OptError::OptError(string o)
{
	opt = o;
	msg = "error processing option: " + opt;
}

OptError::OptError(string o, string m)
{
	opt = o;
	msg = m + " for option:" + o;
}

ConvError::ConvError(string str, string type)
{
	msg = "error converting '" + str + "' to " + type;
}

UnknError::UnknError(string o)
{
	msg = "unknown option: " + o;
}

SetValue::SetValue(int * v)
{
	var = v;
}

void SetValue::add(string name, int value)
{
	set_list[name] = value;
}

void SetValue::set(string str)
{
	if (set_list.find(str) == set_list.end()) arg::Error("unknown value '" + str + "'in set");
	* var = set_list[str];
}
