/* ccgo: go/nc.cc
 *
 * Copyright 2004 Chun-Chung Chen <cjj@u.washington.edu>
 *
 * This program is free software; you can redistribute it and'or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "nc.hh"
extern "C" {
#include <ncurses.h>
}

#include <sstream>
#include <iomanip>
#include <iostream>

using namespace Nc;
using namespace std;

namespace {
	string loc_to_str(int x, int y)
	{
		char c = 'a' + x;
		if (c >= 'i') c ++;
		ostringstream s;
		s << c << y + 1;
		return s.str();
	}
}

// members of class Entry
Entry::Entry() :
	shown(false),
	entx(0),
	enty(0),
	entw(0),
	enth(0),
	entl(0),
	entc(0),
	curl(0),
	curc(0),
	top(0),
	topc(0),
	show_guide(false)
{
	buf.push_back("");
}

Entry::~Entry()
{
}

void Entry::set_size(int x, int y, int width, int height)
{
	entx = x;
	enty = y;
	entw = width;
	enth = height;
	entl = 0;
	entc = 0;
	curl = 0;
	curc = 0;
	top = 0;
	topc = 0;
	shown = true;
	draw_buf();
}

void Entry::draw_buf()
{
	if (! shown) return;
	unsigned bl = top;
	unsigned bc = topc;
	for (int l = 0; l < enth; l ++) {
		int c = 0;
		move(enty + l, entx - 1);
		if (show_guide) {
			if (bc == 0) {
				if (bl < buf.size()) addch(' ');
				else addch('~');
			} else addch('>');
		} else addch(' ');
		if (bl < buf.size()) {
			string s = buf[bl].substr(bc, entw);
			addstr(s.c_str());
			if (int(s.length()) < entw) {
				bl ++;
				bc = 0;
				c += s.length();
			} else {
				bc += s.length();
				c = entw;
			}
		}
		while (c < entw) {
			addch(' ');
			c ++;
		}
	}
}

void Entry::set_cursor()
{
	if (! shown) return;
	// count lines
	entl = 0;
	for (int i = top; i < curl; i ++) {
		entl += buf[i].length() / entw;
		entl ++;
	}
	entl += curc / entw;
	entl -= topc / entw;
	entc = curc % entw;
	move(enty + entl, entx + entc);
}

bool Entry::process_key(int c)
{
	int i;
	bool need_redraw = false;
	switch(c) {
	case KEY_UP:
		if (curc >= entw) {
			curc -= entw;
			if (curl == top && curc < topc) {
				topc -= entw;
				need_redraw = true;
			}
		} else if (curl > 0) {
			curl --;
			curc += entw * (buf[curl].length() / entw);
			if (curc > (int) buf[curl].length()) {
				curc = buf[curl].length();
			}
			if (curl < top) {
				top = curl;
				need_redraw = true;
			}
		}
		break;
	case KEY_LEFT:
		if (curc == 0) break;
		curc --;
		if (curl == top && curc < topc) { // need to scroll?
			topc -= entw;
			need_redraw = true;
		}
		break;
	case KEY_RIGHT:
		if (curc >= (int) buf[curl].length()) break;
		curc ++;
		if (curc % entw == 0 && entl == enth - 1) { // need to scroll?
			topc += entw;
			if (topc > int(buf[top].length())) {
				topc = 0;
				top ++;
			}
			need_redraw = true;
		}
		break;
	case KEY_DOWN:
		curc += entw;
		if (curc >= (int) buf[curl].length()) {
			if (curl >= (int) buf.size() - 1) { // no more lines
				curc -= entw;
				break;
			}
			curc = curc % entw;
			curl ++;
			if (curc > (int) buf[curl].length()) {
				curc = buf[curl].length();
			}
		}
		if (entl == enth - 1) { // need to scroll
			topc += entw;
			if (topc > (int) buf[top].length()) {
				topc = 0;
				top ++;
			}
			need_redraw = true;
		}
		break;
	case 1: // control-a
	case KEY_HOME:
		curc = 0;
		if (curl == top && topc > 0) {
			topc = 0;
			need_redraw = true;
		}
		break;
	case 5: // control-e
	case KEY_END:
		i = buf[curl].length() / entw - curc / entw;
		if (i + entl >= enth) { // need to scroll
			i = i + entl - enth + 1;
			while (i > 0) {
				topc += entw;
				if (topc > (int) buf[top].length()) {
					top ++;
					topc = 0;
				}
				i --;
			}
			need_redraw = true;
		}
		curc = buf[curl].length();
		break;
	case '\r':
		buf.insert(buf.begin() + curl + 1, buf[curl].substr(curc));
		buf[curl].erase(curc);
		curl ++;
		curc = 0;
		if (entl == enth - 1) { // need to scrool
			topc += entw;
			if (topc > (int) buf[top].length()) {
				topc = 0;
				top ++;
			}
		}
		need_redraw = true;
		break;
	case 8: // control-h
	case KEY_BACKSPACE:
		if (curc > 0) {
			curc --;
			buf[curl].erase(curc, 1);
			if (curl == top && curc < topc) {
				topc --;
			}
			need_redraw = true;
		} else if (curl > 0) {
			curl --;
			curc = buf[curl].length();
			buf[curl].append(buf[curl + 1]);
			buf.erase(buf.begin() + curl + 1);
			if (curl < top) {
				top = curl;
				topc = entw * (curc / entw);
			}
			need_redraw = true;
		}
		break;
	case KEY_DC:
	case 4: // control-d
		if (curc < (int) buf[curl].length()) {
			buf[curl].erase(curc, 1);
			need_redraw = true;
		} else if (curl < (int) buf.size() - 1) {
			buf[curl].append(buf[curl + 1]);
			buf.erase(buf.begin() + curl + 1);
			need_redraw = true;
		}
		break;
	case KEY_PPAGE:
		i = enth - 2;
		while (i > 0) {
			if (topc > 0) topc -= entw;
			else if (top > 0) {
				top --;
				topc = entw * (buf[top].length() / entw);
			} else break;
			i --;
			entl ++;
			need_redraw = true;
		}
		while (entl >= enth) {
			if (curc >= entw) curc -= entw;
			else {
				curl --;
				curc = entw * (buf[curl].length() / entw) + curc % entw;
			}
			entl --;
		}
		if (curc > (int) buf[curl].length()) curc = buf[curl].length();
		break;
	case KEY_NPAGE:
		i = enth - 2;
		while (i > 0) {
			if (topc + entw <= (int) buf[top].length()) topc += entw;
			else if (top < (int) buf.size() - 1) {
				top ++;
				topc = 0;
			} else break;
			i --;
			need_redraw = true;
		}
		while (curl < top) {
			curl ++;
			curc = curc % entw;
		}
		while (curc < topc) curc += entw;
		if (curc > (int) buf[curl].length()) curc = buf[curl].length();
		break;
	default:
		if (c >= 32 && c < 128) { // normal key?
			char b[2];
			b[1] = 0;
			b[0] = c;
			buf[curl].insert(curc, b);
			curc ++;
			if (curc % entw == 0 && entl == enth - 1) { // need to scroll?
				topc += entw;
				if (topc > (int) buf[top].length()) {
					topc = 0;
					top ++;
				}
			}
			need_redraw = true;
			break;
		}
		return false;
	}
	if (need_redraw) draw_buf();
	return true;
}

string Entry::get_text()
{
	string t;
	for (vector<string>::const_iterator i = buf.begin(); i != buf.end(); i ++) {
		if (i != buf.begin()) t += '\n';
		t += * i;		
	}
	return t;
}

void Entry::set_text(const string & text)
{
	buf.clear();
	unsigned i = 0;
	unsigned j;
	while ((j = text.find('\n', i)) != string::npos) {
		buf.push_back(text.substr(i, j - i));
		i = j + 1;
	};
	buf.push_back(text.substr(i));
	entl = 0;
	entc = 0;
	curl = 0;
	curc = 0;
	top = 0;
	topc = 0;
}

void Entry::set_guide(bool guide_shown)
{
	show_guide = guide_shown;
}

// members of class Walk
Walk::Walk() :
	board_shown(false),
	status_shown(false),
	map_shown(false),
	comment_shown(false),
	edit_comment(false),
	cy(0),
	cx(0),
	g(0),
	handi(false)
{
	if (cnt == 0) {
		initscr();
		keypad(stdscr, true);
		nonl();
		cbreak();
		noecho();
		if (has_colors()) {
			start_color();
			use_default_colors();
		}
	}
	cnt ++;
	sz = 0;
	resize();
}

Walk::~Walk()
{
	cnt --;
	if (cnt == 0) endwin();
	if (g) delete g;
}

void Walk::set_node(ccgo::Node * node)
{
	ccgo::PathWalk::set_node(node);
	sz = get_size();
	if (sz > 4) {
		cx = 3;
		cy = 3;
	} else {
		cx = 0;
		cy = 0;
	}
	resize();
	for (int i = loc_range() - 1; i >= 0; i --) update_loc(i);
}

bool Walk::process_key(int c)
{
	if (c == KEY_RESIZE) {
		resize();
		return true;
	}
	clear_message();
	if (edit_comment) {
		switch (c) {
		case 3: // control-c
			edit_comment = false;
			show_comment();
			break;
		case '\t':
		case KEY_BTAB:
		case 19: { // control-x
			edit_comment = false;
			string s = te.get_text();
			ccgo::Comment * c = (ccgo::Comment *)(where()->get_prop(ccgo::Comment::the_id()));
			if (s == "") {
				if (c) where()->set_prop(ccgo::Comment::the_id(), 0);
				break;
			}
			if (! c) {
				c = new ccgo::Comment;
				where()->set_prop(ccgo::Comment::the_id(), c);
			}
			c->set_comment(s);
		}
			break;
		default:
			if (! te.process_key(c)) {
				message("unprocessed key!");
			}
		}
		if (! edit_comment) {
			te.set_guide(false);
			te.draw_buf();
		}
		return true;
	}
	int i;
	bool r = true;
	bool node_change = false;
	hide_mark();
	switch (c) {
	case KEY_UP: // move up
		if (cy < sz - 1) cy ++;
		break;
	case KEY_DOWN: // move down
		if (cy > 0) cy --;
		break;
	case KEY_RIGHT: // move right
		if (cx < sz - 1) cx ++;
		break;
	case KEY_LEFT: // move left
		if (cx > 0) cx --;
		break;
	case '\r': // put a stone
		if (board_shown) {
			ccgo::Node * n = make_move(cx + cy * sz);
			if (! n) {
				ostringstream s;
				s << "Move not allowed: " << ::loc_to_str(cx, cy);
				message(s.str());
				break;
			}
			where()->add_child(n);
			PathWalk::down(n);
			node_change = true;
		}
		break;
	case ('d' & 0x1f): // delete a node Ctrl-d
		if (game_del_node()) {
			node_change = true;
		}
		break;
	case 'u':
		if (where()->up()) {
			PathWalk::up();
			node_change = true;
		}
		break;
	case 'd':
		if (! where()->down_num()) break;
		PathWalk::down();
		node_change = true;
		break;
	case '<':
		if (where()->get_level()) {
			unsigned i = where()->get_index();
			if (i < 1) break;
			PathWalk::up();
			PathWalk::down(i - 1);
			node_change = true;
		}
		break;
	case '>':
		if (where()->get_level()) {
			unsigned i = where()->get_index();
			if (i + 1 >= where()->up()->down_num()) break;
			PathWalk::up();
			PathWalk::down(i + 1);
			node_change = true;
		}
		break;
	case 'h': // setup handicap
		if (! game_can_init()) break;
		if (handi) { // unset handi
			game_set_handi(- 1);
			handi = false;
			node_change = true;
			break;
		}
		message("handicap (0,2-9):");
		i = getch();
		clear_message();
		if (i != ERR) {
			int h = i - '0';
			if (h && (h < 2 || h > 9)) {
				message("invalid handicap");
				break;
			}
			try {
				game_set_handi(h);
			} catch (ccgo::Error e) {
				message(string("error: ") + e.get_message());
				break;
			}
			handi = true;
			node_change = true;
		}
		break;
	case 'a': // add handicap stone
		if (! (board_shown && handi && game_can_init())) break;
		game_add_handi(cx + cy * sz);
		node_change = true;
		break;
	case 'p':
		if (board_shown) {
			ccgo::Node * n = make_move(ccgo::MOVE_PASS);
			if (! n) {
				message("Pass not allowed");
				break;
			}
			where()->add_child(n);
			PathWalk::down(n);
			node_change = true;
		}
		break;
	case '\t':
	case KEY_BTAB:
		if (comment_shown) {
			edit_comment = true;
			te.set_guide(true);
			te.draw_buf();
		}
		break;
	case ERR:
		break;
	default:
		r = false;
	}
	if (node_change) {
		show_status();
		show_map();
		show_comment();
		for (std::set<int>::iterator i = ko.begin(); i != ko.end(); i ++) update_loc(* i);
		ko = where()->get_ko();
		for (std::set<int>::iterator i = ko.begin(); i != ko.end(); i ++) update_loc(* i);
	}
	show_mark();
	return r;
}

void Walk::loop()
{
	while (! where()) { // not attached to a game
		message("n)ew game, l)oad game or q)uit:");
		int i = getch();
		switch (i) {
		case 'n':
			g = new ccgo::Game;
			g->set_init(19);
			set_node(g);
			break;
		case 'l':
			break;
		case 'q':
			return;
		}
	}
	do {
		if (edit_comment) te.set_cursor();
		else move(my - 1, 0);

		int c = getch();
		if (process_key(c)) continue;
		if (c == 'q') break;
	} while (true);
}

void Walk::resize()
{
	getmaxyx(stdscr, my, mx);
	if (! where()) {
		return;
	}
	board_shown = my >= sz + 3 && mx >= sz * 2 + 5;  // show only enough size
	status_shown = my >= 5 && mx > sz * 2 + 14;
	map_shown = my > 5 && mx > sz * 2 + 14;
	comment_shown = board_shown && my > sz + 3;
	::clear();
	int xw = mx - (status_shown ? 10 : 0);
	if (xw > 19 * 2 + 6) xw = 19 * 2 + 6;
	bxs = (xw - sz * 2 - 6) / 2;
	if (bxs < 0) bxs = 0;
	sxs = sz * 2 + 6;
	if (sxs < mx - 9) sxs = mx - 9;
	if (sxs > 19 * 2 + 6) sxs = 19 * 2 + 6;
	show_board();
	show_mark();
	show_status();
	show_map();
	if (comment_shown) te.set_size(1, sz + 2, mx - 1, my - sz - 3);
	if (edit_comment) {
		if (comment_shown) te.draw_buf();
		else edit_comment = false; // abort editing mode
	}
	else show_comment();
	er = false;
}

void Walk::update_loc(int loc)
{
	if (! board_shown) return;
	int x = loc % sz;
	int y = loc / sz;
	move(sz - y, bxs + x * 2 + 3);
	switch (get_state(loc)) {
	case STATE_EMPTY:
		if (where()->get_ko().count(loc)) addch('x');
		else if (loc_is_star(loc)) addch('+');
		else addch('.');
		break;
	case STATE_BLACK:
		addch('#');
		break;
	case STATE_WHITE:
		addch('O');
		break;
	}
}

void Walk::show_board()
{
	if (! board_shown) {
		if (sz) mvaddstr(0, 0, "window too small");
		return;
	}

	for (int i = 0; i < sz * sz; i ++) update_loc(i);
	// coordinate system
	for (int i = 0; i < sz; i ++) {
		int c = 'a' + i;
		if (c >= 'i') c ++;
		mvaddch(sz + 1, bxs + i * 2 + 3, c);
		mvaddch(0, bxs + i * 2 + 3, c);
		c = '0' + (i + 1) % 10;
		mvaddch(sz - i, bxs + 1, c);
		mvaddch(sz - i, bxs + sz * 2 + 3 + (i + 1) / 10, c);
		if (i > 8) {
			c = '0' + (i + 1) / 10;
			mvaddch(sz - i, bxs, c);
			mvaddch(sz - i, bxs + sz * 2 + 3, c);
		}
	}
}

void Walk::show_mark()
{
	if (! board_shown) return;
	mvaddch(sz - cy, bxs + cx * 2 + 2, '[');
	mvaddch(sz - cy, bxs + cx * 2 + 4, ']');
}

void Walk::hide_mark()
{
	if (er) {
		move(my - 1, 1);
		addstr("                                    ");
		er = false;
	}
	if (! board_shown) return;
	mvaddch(sz - cy, bxs + cx * 2 + 2, ' ');
	mvaddch(sz - cy, bxs + cx * 2 + 4, ' ');
}

void Walk::show_status()
{
	if (! status_shown) return;
	int l = my > sz + 3 ? sz - 2 : my - 5;
	mvaddstr(l, sxs, "=========");
	ostringstream s;
	s << "Move " << setw(3) << left << where()->get_num();
	mvaddstr(++ l, sxs + 1, s.str().c_str());
	s.str("");
	s << "Turn: " << (get_turn() == Position::TURN_BLACK ? '#' : 'O');
	mvaddstr(++ l, sxs + 1, s.str().c_str());
	s.str("");
	s << "Alt: ";
	if (where()->get_level()) {
		unsigned i = where()->get_index();
		s << (i > 0 ? '<' : ' ');
		s << '-';
		s << (i + 1 < where()->up()->down_num() ? '>' : ' ');
	} else s << " - ";
	mvaddstr(++ l, sxs + 1, s.str().c_str());

	int m = where()->get_move();
	if (m >= 0) {
		hide_mark();
		cx = m % sz;
		cy = m / sz;
		show_mark();
	}
}

void Walk::show_map()
{
	if (! map_shown) return;
	int l = my > sz + 3 ? sz - 2 : my - 5;
	int c = where()->get_level();
	// find last node
	ccgo::Node * n = where();
	int k = path_depth();
	int s = k < l || c < l / 2 ? 0 : c + l > k + l / 2 + 1 ? k + 1 - l : c - l / 2;
	// find beginning node
	n = where();
	k = c;
	ccgo::Position::Turn u = get_turn(); // keep track of the turn played
	while (k > s) {
		if (n->get_delta() && n->get_delta()->get_turn_flip()) u = ccgo::Position::flip_turn(u);
		n = n->up();
		k --;
	}
	if (n->get_delta() && n->get_delta()->get_turn_flip()) u = ccgo::Position::flip_turn(u);
	for (int i = 0; i < l; i ++) {
		ostringstream t;
		if (n) {
			bool a1 = false;
			bool a2 = false;
			if (n->get_level()) {
				unsigned j = n->get_index();
				if (j > 0) a1 = true;
				if (j + 1 < n->up()->down_num()) a2 = true;
			}
			t << (k == c ? '[' : ' ');
			t << (a1 ? '<' : ' ');
			int m = n->get_move();
			if (m == ccgo::MOVE_NONE) {
				if (k) t << " --- "; // no move
				else t << (handi ? "handi" : "start");
			} else {
				t << (u == ccgo::Position::TURN_BLACK ? "#:" : "O:");
				if (m == ccgo::MOVE_PASS) t << "pa.";
				else t << setw(3) << left << ::loc_to_str(m % sz, m / sz);
			}
			t << (a2 ? '>' : ' ');
			t << (k == c ? ']' : ' ');
			if (n->get_delta() && n->get_delta()->get_turn_flip()) u = ccgo::Position::flip_turn(u);
			if (k < (int) path_depth()) n = get_path(k);
			else if (n->down_num()) n = n->down(0);
			else n = 0;
			k ++;
		} else t << "         ";
		mvaddstr(i, sxs, t.str().c_str());
	}
}

void Walk::show_comment()
{
	if (! comment_shown) return;
	ccgo::Comment * c = (ccgo::Comment *)(where()->get_prop(ccgo::Comment::the_id()));
	if (c) te.set_text(c->get_comment());
	else te.set_text("");
	te.set_guide(false);
	te.draw_buf();
}

void Walk::clear()
{
	ccgo::PathWalk::clear();
	for (int i = loc_range() - 1; i >= 0; i --) update_loc(i);
}

void Walk::set_state(int loc, State state)
{
	ccgo::PathWalk::set_state(loc, state);
	update_loc(loc);
}

void Walk::clear_message()
{
	move(my - 1, 1);
	clrtoeol();
	er = false;
}

void Walk::message(const string & s)
{
	if (er)	clear_message();
	mvaddstr(my - 1, 1, s.c_str());
	er = true;
}

int Walk::cnt = 0;
